/**range [0,1) */
function urandom()
{return Math.random()}
/**uniform distribution [-scale,scale)*/
function uni(scale=1)
{return (urandom()-0.5)*2*scale}
/** 
 * @template T
 * @param {T[]} arr
 * @returns {T}
 * */
function choose(arr=[])
{
    if (Array.isArray(arr))
    {
        return arr[parseInt(urandom()*arr.length)]
    }
    return arr
}

/** return index of minimum
 * @template T
 * @param {T[]} arr
 * @param {(x:T)=>Number} key
 * @returns {T}
 * */
function argmin(arr=[],key=parseInt)
{
    var k=arr.map(x=>key(x)).map(x=>isNaN(x)?Infinity:x)
    var kmin=Math.min(...(k.flat()))
    return k.findIndex(x=>x==kmin)
}
/**
 * @template T
 * @param {T[]} arr
 * @param {(x:T)=>Number} key
 * @returns {T}
 */
function min(arr=[],key=parseInt,_default=undefined)
{
    if (arr.length==0 && _default!==undefined)return _default
    return arr[argmin(arr,key)]
}
/**
 * @template T
 * @param {T[]} arr
 * @param {(x:T)=>Number} key
 * @returns {T}
 */
function max(arr=[],key=parseInt,_default=undefined)
{
    return min(arr,i=>-key(i),_default)
}
/** return index，from minimum's to maximum's
 * @template T
 * @param {T[]} arr
 * @param {(x:T)=>Number} key
 * @returns {Number[]}
 * */
function argsort(arr=[],key=parseInt)
{
    var k=arr.map((x,i)=>key(x)).map(x=>isNaN(x)?Infinity:x).map((x,i)=>[x,i])
    k.sort((a,b)=>(a[0]-b[0]))
    return k.map(a=>a[1])
}

/** return index in sorted array
 * @template T
 * @param {T[]} arr
 * @param {(x:T)=>Number} key
 * @returns {Number[]}
 * */
function argmap(arr=[],key=parseInt)
{
    var arr2=argsort(arr,key)
    var arr3=new Array(arr.length)
    arr2.forEach((i,j)=>arr3[i]=j)
    return arr3
}

/**
 * @param {Number[]} arr
 */
function sum(arr=[])
{return arr.map(x=>isNaN(x)?0:x).flat().reduce((a,b)=>(a+b))}


/**
 * @template T
 * @param {T[]} arr1
 * @param {T[]} arr2
 * @param {(a:T,b:T)=>Number} key
 * @returns {T[][]}
 */
function pair(arr1=[],arr2=[],key=(a,b)=>Math.abs(a-b))
{
    var arr1=Array.from(arr1).reverse()
    var arr2=Array.from(arr2)
    var out=[]
    while (arr1.length>0 && arr2.length>0)
    {
        var a=arr1.pop()
        var bi=argmin(arr2,(b)=>key(a,b))
        var b=arr2.splice(bi,1)[0]
        out.push([a,b])
    }
    return out
}
/**
 * @template T
 * @param {T[]} arr
 * @param {(a:T)=>Number} key
 * @returns {T[][]}
 */
function groupBy(arr=[],key=i=>0)
{
    var group=[]
    arr.forEach((i)=>{
        var k=key(i)
        if (group[k]===undefined)group[k]=[]
        group[k].push(i)
    })
    return group
}





/**
 * @param {Number} ratio
 * @param {_object} a
 * @param {_object} b
 */
function mix2fv(ratio,a,b)
{
    return {
        x: a.x*ratio + b.x*(1-ratio),
        y: a.y*ratio + b.y*(1-ratio)
    }
}
















function distBetween(object,object2)
{
    if (object===undefined || object2===undefined)return Infinity
    return distBetweenTwoPoints(object.x,object.y,object2.x,object2.y)
}

function distTo(object)
{return (object2)=>distBetween(object,object2)}

function enumOil()
{return enumFeature(ALL_PLAYERS).filter((x)=>(x.stattype === OIL_RESOURCE))}
/** 
 * @returns {String[]}
*/
function enumStat(x=Stats.Building)
{return Object.keys(x).map(i=>x[i].Id)}

function enumEnemy()
{
    return startPositions.map((x,i)=>i).filter(i=>!allianceExistsBetween(me,i))
}

/** build as a larger building will preserve more space */
function spamBuild(truck,struct="A0PowerGenerator",near,as,spread=0,safe=true)
{
    if (near===undefined)near=truck
    if (as===undefined)as=struct
    var pos=pickStructLocation(truck,as,near.x+uni(spread),near.y+uni(spread))
    if (struct=="A0ResourceExtractor")pos=near
    if (pos===undefined)return
    if (safe && !safeDest(me,pos.x,pos.y))return
    return orderDroidBuild(truck, DORDER_BUILD, struct, pos.x,pos.y)
}

function spamMove(droid,scale=5,near,order=DORDER_MOVE,range=Infinity)
{
    if (near===undefined)near=droid
    if (range!==Infinity)
    {
        var d=Math.hypot(near.x-droid.x,near.y-droid.y)
        var coef=range/d
        if (d>range)near={"x":droid.x+(near.x-droid.x)*coef,"y":droid.y+(near.y-droid.y)*coef}
    }
    try{orderDroidLoc(droid,order,near.x+Math.round(uni(scale)),near.y+Math.round(uni(scale)))}
    catch{}
}



